﻿using HigLabo.Data;
using HigLabo.DbSharp.CodeGenerator;
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading;
using System.Threading.Tasks;
using System.Windows.Threading;

namespace HigLabo.DbSharpApplication.Core
{
    public class ImportTableProcess
    {
        public event EventHandler<ProcessProgressEventArgs> Executed;
        public event EventHandler Completed;
        private List<Table> Tables { get; set; }
        private List<StoredProcedure> StoredProcedures { get; set; }
        public Dispatcher Dispatcher { get; private set; }
        public DatabaseSchemaReader DatabaseSchemaReader { get; private set; }
        public ImportTableProcess(Dispatcher dispatcher, DatabaseServer databaseServer, String connectionString)
        {
            if (dispatcher == null) throw new ArgumentNullException("dispatcher");

            this.Dispatcher = dispatcher;
            this.DatabaseSchemaReader = DatabaseSchemaReader.Create(databaseServer, connectionString);
            this.Tables = new List<Table>();
            this.StoredProcedures = new List<StoredProcedure>();
        }
        public void Start(List<String> names)
        {
            ThreadPool.QueueUserWorkItem(o => this.Execute(names));
        }
        private void Execute(List<String> names)
        {
            DatabaseSchemaReader r = this.DatabaseSchemaReader;
            var totalCount = names.Count;

            this.Tables.Clear();
            this.StoredProcedures.Clear();
            for (int i = 0; i < totalCount; i++)
            {
                var name = names[i];
                var t = r.GetTable(name);

                var tExisted = AValue.SchemaData.Tables.FirstOrDefault(el => el.Name == t.Name);
                if (tExisted == null)
                {
                    this.Tables.Add(t);
                }
                else if (tExisted.LastAlteredTime < t.LastAlteredTime)//DB上に定義されているテーブルが更新されている場合
                {
                    foreach (var parameter in t.Columns)
                    {
                        var p = tExisted.Columns.Find(el => el.Name == parameter.Name);
                        if (p == null) { continue; }
                        //手動設定の値をコピー
                        parameter.EnumName = p.EnumName;
                    }
                    this.Tables.Add(t);
                }
                //CRUD用のストアドをDBに追加
                this.AddStoredProcedure(t);
                this.DispatcherInvoke(() => this.OnExecuted(new ProcessProgressEventArgs(t.Name, i / totalCount)));
            }
            //スキーマファイルに追加
            this.DispatcherInvoke(this.AddToSchemaFile);
            this.DispatcherInvoke(this.OnCompleted);
        }
        private void DispatcherInvoke(Action action)
        {
            this.Dispatcher.Invoke(action);
        }
        public void AddStoredProcedure(Table table)
        {
            var r = this.DatabaseSchemaReader;

            using (var db = this.DatabaseSchemaReader.CreateDatabase())
            {
                this.AddStoredProcedure(db, table, table.Name + "SelectAll", r.CreateQueryOfTableNameSelectAll(table));
                this.AddStoredProcedure(db, table, table.Name + "SelectBy_PrimaryKey", r.CreateQueryOfTableNameSelectBy_PrimaryKey(table));
                this.AddStoredProcedure(db, table, table.Name + "Insert", r.CreateQueryOfTableNameInsert(table));
                this.AddStoredProcedure(db, table, table.Name + "Update", r.CreateQueryOfTableNameUpdate(table));
                this.AddStoredProcedure(db, table, table.Name + "Delete", r.CreateQueryOfTableNameDelete(table));
            }
        }
        private void AddStoredProcedure(IDatabase database, Table table, String storedProcedureName, String query)
        {
            var r = this.DatabaseSchemaReader;
            var db = database;

            if (r.ExistStoredProcedure(storedProcedureName) == true)
            {
                db.ExecuteCommand("drop procedure " + storedProcedureName, true);
            }
            db.ExecuteCommand(query, true);
            var sp = r.GetStoredProcedure(storedProcedureName);
            this.SetSchema(table, sp);
            this.StoredProcedures.Add(sp);
        }
        private void SetSchema(Table table, StoredProcedure storedProcedure)
        {
            var sp = storedProcedure;
            Column c = null;
            String pName = "";

            foreach (var p in sp.Parameters)
            {
                pName = p.GetNameWithoutAtmark();
                if (pName.StartsWith("PK_") == true)
                {
                    pName = pName.Substring(3, pName.Length - 3);
                }
                c = table.Columns.Find(el => el.Name == pName);
                if (c == null) { continue; }
                p.AllowNull = c.AllowNull;
                p.EnumName = c.EnumName;
            }
            foreach (var p in sp.ResultSets)
            {
                pName = p.GetNameWithoutAtmark();
                if (pName.StartsWith("PK_") == true)
                {
                    pName = pName.Substring(3, pName.Length - 3);
                }
                c = table.Columns.Find(el => el.Name == pName);
                if (c == null) { continue; }
                p.AllowNull = c.AllowNull;
                p.EnumName = c.EnumName;
            }
            sp.TableName = table.Name;
        }
        private void AddToSchemaFile()
        {
            foreach (var table in this.Tables)
            {
                var t = AValue.SchemaData.Tables.FirstOrDefault(el => el.Name == table.Name);
                if (t != null)
                {
                    AValue.SchemaData.Tables.Remove(t);
                }
                AValue.SchemaData.Tables.Add(table);
            }
            foreach (var sp in this.StoredProcedures)
            {
                var spExist = AValue.SchemaData.StoredProcedures.FirstOrDefault(el => el.Name == sp.Name);
                if (spExist != null)
                {
                    AValue.SchemaData.StoredProcedures.Remove(spExist);
                }
                AValue.SchemaData.StoredProcedures.Add(sp);
            }
            AValue.SchemaData.LastExecuteTimeOfImportTable = DateTime.Now;
        }
        private void OnExecuted(ProcessProgressEventArgs e)
        {
            var eh = this.Executed;
            if (eh != null)
            {
                eh(this, e);
            }
        }
        private void OnCompleted()
        {
            var eh = this.Completed;
            if (eh != null)
            {
                eh(this, EventArgs.Empty);
            }
        }
    }
}
