﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using HigLabo.DbSharp;
using HigLabo.CodeGenerator;
using System.Data;
using System.IO;

namespace HigLabo.DbSharp.CodeGenerator
{
    public class StoredProcedureClassSourceCodeFileFactory : ClassSourceCodeFileFactory
    {
        public StoredProcedure StoredProcedure { get; private set; }

        public StoredProcedureClassSourceCodeFileFactory(StoredProcedure storedProcedure)
        {
            this.StoredProcedure = storedProcedure;
        }
        public override SourceCodeFile Create(String directoryPath, String namespaceName)
        {
            var sp = this.StoredProcedure;
            var sc = new SourceCode();

            sc.UsingNamespaces.Add("System");
            sc.UsingNamespaces.Add("System.Data");
            sc.UsingNamespaces.Add("System.Text");
            sc.UsingNamespaces.Add("HigLabo.DbSharp");

            var ns = new Namespace(namespaceName);
            ns.Classes.Add(CreateClass());
            sc.Namespaces.Add(ns);

            return new SourceCodeFile(Path.Combine(directoryPath, sp.Name + ".cs"), sc);
        }
     
        public Class CreateClass()
        {
            var sp = this.StoredProcedure;
            Class c = new Class(AccessModifier.Public, sp.Name);
            Boolean hasResultSet = false;

            c.Modifier.Partial = true;
            if (sp.ResultSets.Count > 0)
            {
                hasResultSet = true;
            }
            if (hasResultSet == true)
            {
                c.BaseClass = new TypeName(String.Format("StoredProcedureWithResultSet<{0}.ResultSet>", sp.Name));
            }
            else
            {
                c.BaseClass = new TypeName("StoredProcedure");
            }
            c.Fields.Add(CreateNameConstField());

            if (sp.Parameters.Exists(el => el.Name == "TransactionKey") == false)
            {
                c.Properties.Add(CreateTransactionKeyProperty());
            }

            ClassSourceCodeFileFactory.AddPropertyAndField(c, sp.Parameters);
            foreach (var parameter in sp.Parameters)
            {
                if (sp.DatabaseServer == DatabaseServer.SqlServer &&
                    parameter.DbType.IsStructured() == true)
                {
                    c.Methods.Add(CreateCreateDataTableMethod(parameter));
                }                
            }

            c.Constructors.Add(CreateConstructor());

            c.Methods.Add(CreateGetStoredProcedureNameMethod());
            c.Methods.Add(new Method(MethodAccessModifier.Partial, "ConstructorExecuted"));
            c.Methods.Add(CreateCreateCommandMethod());
            c.Methods.Add(CreateSetOutputParameterValueMethod());

            if (hasResultSet == true)
            {
                c.Methods.Add(CreateCreateResultSetMethod());
                c.Methods.Add(CreateSetResultSetMethod());
                c.Methods.Add(CreateGetResultSetMethod());

                var f = new ResultSetClassFactory(sp);
                c.Classes.Add(f.CreateClass());
            }
            c.Methods.Add(CreateToStringMethod());

            return c;
        }

        public Field CreateNameConstField()
        {
            var f = new Field("String", "Name");
            f.Modifier.AccessModifier = AccessModifier.Public;
            f.Modifier.Const = true;
            f.Initializer = String.Format("\"{0}\"", this.StoredProcedure.Name);
            return f;
        }
        public Property CreateTransactionKeyProperty()
        {
            Property p = new Property("String", "TransactionKey");
            p.Modifier.AccessModifier = MethodAccessModifier.Public;
            p.Get.Body.Add(SourceCodeLanguage.CSharp, "return ((ITransaction)this).TransactionKey;");
            p.Set.Body.Add(SourceCodeLanguage.CSharp, "((ITransaction)this).TransactionKey = value;");

            return p;
        }
        
        public Method CreateGetStoredProcedureNameMethod()
        {
            var sp = this.StoredProcedure;
            var md = new Method( MethodAccessModifier.Public, "GetStoredProcedureName");
            md.Modifier.Polymophism = MethodPolymophism.Override;
            md.ReturnTypeName = new TypeName("String");
            md.Body.Add(SourceCodeLanguage.CSharp, "return {0}.Name;", sp.Name);

            return md;
        }
  
        private static Method CreateCreateDataTableMethod(StoredProcedureParameter parameter)
        {
            Method md = new Method(MethodAccessModifier.Private, String.Format("Create{0}DataTable", parameter.GetNameWithoutAtmark()));
            md.ReturnTypeName = new TypeName("DataTable");
            md.Body.AddRange(CreateDataTableMethodBody(parameter));
            return md;
        }
        private static IEnumerable<CodeBlock> CreateDataTableMethodBody(StoredProcedureParameter parameter)
        {
            yield return new CodeBlock(SourceCodeLanguage.CSharp,"var dt = new DataTable();");
            foreach (var column in parameter.StructuredDataTableColumns)
            {
                yield return new CodeBlock(SourceCodeLanguage.CSharp, "dt.Columns.Add(\"{0}\", typeof({1}));", column.ColumnName, column.ClassNameType.ToClassNameString());
            }
            yield return new CodeBlock(SourceCodeLanguage.CSharp, "return dt;");
        }

        public Constructor CreateConstructor()
        {
            var sp = this.StoredProcedure;
            Constructor ct = new Constructor(AccessModifier.Public, sp.Name);

            foreach (var parameter in sp.Parameters)
            {
                if (sp.DatabaseServer == DatabaseServer.SqlServer &&
                    parameter.DbType.IsStructured() == true)
                {
                    ct.Body.Add(new CodeBlock(SourceCodeLanguage.CSharp, "this.{0} = this.Create{0}DataTable();", parameter.GetNameWithoutAtmark()));
                }
            }
            ct.Body.Add(SourceCodeLanguage.CSharp, "ConstructorExecuted();");
            return ct;
        }

        public Method CreateCreateCommandMethod()
        {
            var sp = this.StoredProcedure;
            Method md = new Method(MethodAccessModifier.Public, "CreateCommand");
            md.Modifier.Polymophism = MethodPolymophism.Override;
            md.ReturnTypeName = new TypeName("IDbCommand");

            md.Body.AddRange(CreateCreateCommandMethodBody());

            return md;
        }
        public IEnumerable<CodeBlock> CreateCreateCommandMethodBody()
        {
            var sp = this.StoredProcedure;
            String pName = "";
            
            switch (sp.DatabaseServer)
            {
                case DatabaseServer.SqlServer: yield return new CodeBlock(SourceCodeLanguage.CSharp, "var cm = new System.Data.SqlClient.SqlCommand();"); break;
                case DatabaseServer.Oracle: yield return new CodeBlock(SourceCodeLanguage.CSharp, "var cm = new Oracle.DataAccess.Client.OracleCommand();"); break;
                case DatabaseServer.MySql: yield return new CodeBlock(SourceCodeLanguage.CSharp, "var cm = new MySql.Data.MySqlClient.MySqlCommand();"); break;
                case DatabaseServer.PostgreSql: yield return new CodeBlock(SourceCodeLanguage.CSharp, "var cm = new Npgsql.NpgsqlCommand();"); break;
                default: throw new InvalidOperationException();
            }
            yield return new CodeBlock(SourceCodeLanguage.CSharp, "cm.CommandType = CommandType.StoredProcedure;");
            yield return new CodeBlock(SourceCodeLanguage.CSharp, "cm.CommandText = \"{0}\";", sp.Name);

            if (sp.Parameters.Count > 0)
            {
                yield return new CodeBlock(SourceCodeLanguage.CSharp, "");
                switch (sp.DatabaseServer)
                {
                    case DatabaseServer.SqlServer: yield return new CodeBlock(SourceCodeLanguage.CSharp, "System.Data.SqlClient.SqlParameter p = null;"); break;
                    case DatabaseServer.Oracle: yield return new CodeBlock(SourceCodeLanguage.CSharp, "Oracle.DataAccess.Client.OracleParameter p = null;"); break;
                    case DatabaseServer.MySql: yield return new CodeBlock(SourceCodeLanguage.CSharp, "MySql.Data.MySqlClient.MySqlParameter p = null;"); break;
                    case DatabaseServer.PostgreSql: yield return new CodeBlock(SourceCodeLanguage.CSharp, "Npgsql.NpgsqlParameter p = null;"); break;
                    default: throw new InvalidOperationException();
                }
                yield return new CodeBlock(SourceCodeLanguage.CSharp, "");
                
                foreach (var parameter in sp.Parameters)
                {
                    pName = parameter.GetNameWithoutAtmark();

                    yield return new CodeBlock(SourceCodeLanguage.CSharp, "p = cm.Parameters.Add(\"{0}\", {1});"
                        , parameter.Name, parameter.DbType.GetDbTypeCodeBlock());
                    yield return new CodeBlock(SourceCodeLanguage.CSharp, "p.SourceColumn = p.ParameterName;");
                    switch (parameter.ParameterDirection)
                    {
                        case ParameterDirection.Input: yield return new CodeBlock(SourceCodeLanguage.CSharp, "p.Direction = ParameterDirection.Input;"); break;
                        case ParameterDirection.InputOutput: yield return new CodeBlock(SourceCodeLanguage.CSharp, "p.Direction = ParameterDirection.InputOutput;"); break;
                        case ParameterDirection.Output: yield return new CodeBlock(SourceCodeLanguage.CSharp, "p.Direction = ParameterDirection.Output;"); break;
                        case ParameterDirection.ReturnValue: yield return new CodeBlock(SourceCodeLanguage.CSharp, "p.Direction = ParameterDirection.ReturnValue;"); break;
                        default: throw new InvalidOperationException();
                    }
                    if (parameter.DbType.IsStructured() == true)
                    {
                        yield return new CodeBlock(SourceCodeLanguage.CSharp, "p.TypeName = \"{0}\";", parameter.UserTableTypeName);
                    }
                    if (parameter.DbType.IsUdt() == true)
                    {
                        yield return new CodeBlock(SourceCodeLanguage.CSharp, "p.UdtTypeName = \"{0}\";", parameter.UdtTypeName);
                    }
                    yield return new CodeBlock(SourceCodeLanguage.CSharp, "p.Size = {0};", parameter.Length);
                    yield return new CodeBlock(SourceCodeLanguage.CSharp, "p.Scale = {0};", parameter.Scale);
                    yield return new CodeBlock(SourceCodeLanguage.CSharp, "p.Precision = {0};", parameter.Precision);
                    switch (parameter.ConvertType)
                    {
                        case SqlParameterConvertType.Default: yield return new CodeBlock(SourceCodeLanguage.CSharp, "p.Value = this.{0};", pName); break;
                        case SqlParameterConvertType.Enum:
                            {
                                if (parameter.AllowNull == true)
                                {
                                    yield return new CodeBlock(SourceCodeLanguage.CSharp, "if (this.{0} != null) p.Value = this.{0}.ToString();", pName); break;
                                }
                                else
                                {
                                    yield return new CodeBlock(SourceCodeLanguage.CSharp, "p.Value = this.{0}.ToString();", pName); break;
                                }
                            }
                        default: throw new InvalidOperationException();
                    }
                    yield return new CodeBlock(SourceCodeLanguage.CSharp, "");
                }
                yield return new CodeBlock(SourceCodeLanguage.CSharp, "");

                var cb = new CodeBlock(SourceCodeLanguage.CSharp, "for (int i = 0; i < cm.Parameters.Count; i++)");
                cb.CurlyBracket = true;
                cb.CodeBlocks.Add(new CodeBlock(SourceCodeLanguage.CSharp, "if (cm.Parameters[i].Value == null) cm.Parameters[i].Value = DBNull.Value;"));
                yield return cb;
            }

            yield return new CodeBlock(SourceCodeLanguage.CSharp, "return cm;");
        }

        public Method CreateSetOutputParameterValueMethod()
        {
            var sp = this.StoredProcedure;
            Method md = new Method(MethodAccessModifier.Protected, "SetOutputParameterValue");
            md.Modifier.Polymophism = MethodPolymophism.Override;
            md.ReturnTypeName = new TypeName("void");

            md.AddParameter("IDbCommand", "command");

            md.Body.AddRange(CreateSetOutputParameterValueMethodBody());

            return md;
        }
        public IEnumerable<CodeBlock> CreateSetOutputParameterValueMethodBody()
        {
            var sp = this.StoredProcedure;
            Boolean addedInitializeCodeBlcok = false;

            foreach (var parameter in sp.Parameters)
            {
                if (parameter.ParameterDirection == System.Data.ParameterDirection.Input) continue;
                if (addedInitializeCodeBlcok == false)
                {
                    yield return new CodeBlock(SourceCodeLanguage.CSharp, "var cm = command;");
                    yield return new CodeBlock(SourceCodeLanguage.CSharp, "IDbDataParameter p = null;");
                    addedInitializeCodeBlcok = true;
                }
                yield return new CodeBlock(SourceCodeLanguage.CSharp, "p = cm.Parameters[{0}] as IDbDataParameter;", parameter.Ordinal);
                switch (parameter.ConvertType)
                {
                    case SqlParameterConvertType.Default: 
                        yield return new CodeBlock(SourceCodeLanguage.CSharp, "if (p.Value != DBNull.Value && p.Value != null) this.{0} = ({1})p.Value;"
                            , parameter.GetNameWithoutAtmark(), parameter.ClassNameType.ToClassNameString());
                        break;
                    case SqlParameterConvertType.Enum:
                        yield return new CodeBlock(SourceCodeLanguage.CSharp, "if (p.Value != DBNull.Value && p.Value != null) this.{0} = StoredProcedure.ToEnum<MyEnum>(p.Value as String) ?? this.{0};"
                            , parameter.GetNameWithoutAtmark(), parameter.EnumName);
                        break;
                    default: throw new InvalidOperationException();
                }
            }
        }

        public Method CreateCreateResultSetMethod()
        {
            var sp = this.StoredProcedure;
            Method md = new Method(MethodAccessModifier.Public, "CreateResultSet");
            md.Modifier.Polymophism = MethodPolymophism.Override;
            md.ReturnTypeName = new TypeName(sp.Name + ".ResultSet");

            md.Body.Add(SourceCodeLanguage.CSharp, "return new ResultSet(this);", sp.Name);
            return md;
        }
        public Method CreateGetResultSetMethod()
        {
            var sp = this.StoredProcedure;
            Method md = new Method(MethodAccessModifier.Public, "GetResultSet");
            md.Modifier.Polymophism = MethodPolymophism.New;
            md.ReturnTypeName = new TypeName(sp.Name + ".ResultSet");

            md.Body.Add(SourceCodeLanguage.CSharp, "return base.GetResultSet() as {0}.ResultSet;", sp.Name);
            return md;
        }
        
        public Method CreateSetResultSetMethod()
        {
            var sp = this.StoredProcedure;
            Method md = new Method(MethodAccessModifier.Protected, "SetResultSet");
            md.Modifier.Polymophism = MethodPolymophism.Override;

            md.AddParameter(sp.Name + ".ResultSet", "resultSet");
            md.AddParameter("IDataReader", "reader");

            md.Body.AddRange(CreateSetResultSetMethodBody());
            return md;
        }
        public IEnumerable<CodeBlock> CreateSetResultSetMethodBody()
        {
            var sp = this.StoredProcedure;
            CodeBlock cb = null;
            String code = "";

            yield return new CodeBlock(SourceCodeLanguage.CSharp, "var r = resultSet;");
            yield return new CodeBlock(SourceCodeLanguage.CSharp, "Int32 index = -1;");

            cb = new CodeBlock(SourceCodeLanguage.CSharp, "try");
            cb.CurlyBracket = true;
            foreach (var parameter in sp.ResultSets)
            {
                code = "index += 1; ";
                if (parameter.AllowNull == true)
                {
                    code += "if (reader[index] != DBNull.Value) ";
                }
                switch (parameter.ConvertType)
                {
                    case SqlParameterConvertType.Default: code += String.Format("r.{0} = {1};", parameter.Name, this.CreateDataReaderCastCSharpCode(parameter.DbType)); break;
                    case SqlParameterConvertType.Enum: code += String.Format("r.{0} = StoredProcedure.ToEnum<{1}>(reader[index] as String) ?? r.{0};", parameter.Name, parameter.EnumName); break;
                    default: throw new InvalidOperationException();
                }
                cb.CodeBlocks.Add(new CodeBlock(SourceCodeLanguage.CSharp, code));
            }
            yield return cb;

            cb = new CodeBlock(SourceCodeLanguage.CSharp, "catch (InvalidCastException ex)");
            cb.CurlyBracket = true;
            cb.CodeBlocks.Add(new CodeBlock(SourceCodeLanguage.CSharp, "throw new StoredProcedureSchemaMismatchedException(this, index, ex);"));
            yield return cb;
        }
        public String CreateDataReaderCastCSharpCode(DbType type)
        {
            switch (type.DatabaseServer)
            {
                case DatabaseServer.SqlServer: return this.CreateDataReaderCastCSharpCode(type.SqlServerDbType.Value);
                case DatabaseServer.Oracle:
                case DatabaseServer.MySql:
                case DatabaseServer.PostgreSql: throw new NotImplementedException();
                default: throw new InvalidOperationException();
            }
        }
        private String CreateDataReaderCastCSharpCode(SqlServer2012DbType type)
        {
            switch (type)
            {
                case SqlServer2012DbType.BigInt:
                    return "reader.GetInt64(index)";

                case SqlServer2012DbType.Image:
                case SqlServer2012DbType.Binary:
                case SqlServer2012DbType.Timestamp:
                case SqlServer2012DbType.VarBinary:
                    return "reader[index] as Byte[]";

                case SqlServer2012DbType.Bit:
                    return "reader.GetBoolean(index)";

                case SqlServer2012DbType.Char:
                case SqlServer2012DbType.NChar:
                case SqlServer2012DbType.NText:
                case SqlServer2012DbType.NVarChar:
                case SqlServer2012DbType.Text:
                case SqlServer2012DbType.VarChar:
                case SqlServer2012DbType.Xml:
                    return "reader[index] as String";

                case SqlServer2012DbType.DateTime:
                case SqlServer2012DbType.SmallDateTime:
                case SqlServer2012DbType.Date:
                case SqlServer2012DbType.DateTime2:
                    return "reader.GetDateTime(index)";

                case SqlServer2012DbType.Time:
                    return "(TimeSpan)reader[index]";

                case SqlServer2012DbType.DateTimeOffset:
                    return "(DateTimeOffset)reader[index]";

                case SqlServer2012DbType.Decimal:
                case SqlServer2012DbType.Money:
                case SqlServer2012DbType.SmallMoney:
                    return "reader.GetDecimal(index)";

                case SqlServer2012DbType.Float:
                    return "reader.GetDouble(index)";

                case SqlServer2012DbType.Int:
                    return "reader.GetInt32(index)";

                case SqlServer2012DbType.Real:
                    return "reader.GetFloat(index)";

                case SqlServer2012DbType.SmallInt:
                    return "reader.GetInt16(index)";

                case SqlServer2012DbType.Structured:
                    return "reader[index] as DataTable";

                case SqlServer2012DbType.TinyInt:
                    return "reader.GetByte(index)";

                case SqlServer2012DbType.Variant:
                case SqlServer2012DbType.Udt:
                    return "reader[index] as Object";

                case SqlServer2012DbType.UniqueIdentifier:
                    return "reader.GetGuid(index)";

                default:
                    return "reader[index] as Object";
            }
        }

        public Method CreateToStringMethod()
        {
            var sp = this.StoredProcedure;
            Method md = new Method(MethodAccessModifier.Public, "ToString");
            md.ReturnTypeName =  new TypeName("String");
            md.Modifier.Polymophism = MethodPolymophism.Override;
            md.Body.AddRange(CreateToStringMethodBody());

            return md;
        }
        public IEnumerable<CodeBlock> CreateToStringMethodBody()
        {
            var sp = this.StoredProcedure;

            yield return new CodeBlock(SourceCodeLanguage.CSharp, "var sb = new StringBuilder(32);");
            yield return new CodeBlock(SourceCodeLanguage.CSharp, "sb.AppendLine(\"<{0}>\");", sp.Name);

            foreach (var parameter in sp.Parameters)
            {
                yield return new CodeBlock(SourceCodeLanguage.CSharp, "sb.AppendFormat(\"{0}={1}\", this.{0}); sb.AppendLine();", parameter.GetNameWithoutAtmark(), "{0}");
            }
            yield return new CodeBlock(SourceCodeLanguage.CSharp, "return sb.ToString();");
        }
    }
}
