﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using HigLabo.CodeGenerator;
using System.IO;

namespace HigLabo.DbSharp.CodeGenerator
{
    public class TableClassSourceCodeFileFactory : ClassSourceCodeFileFactory
    {
        public Table Table { get; private set; }

        public TableClassSourceCodeFileFactory(Table table)
        {
            this.Table = table;
        }
        public override SourceCodeFile Create(String directoryPath, String namespaceName)
        {
            var t = this.Table;
            var sc = new SourceCode();

            sc.UsingNamespaces.Add("System");
            sc.UsingNamespaces.Add("System.Linq");
            sc.UsingNamespaces.Add("System.Data");
            sc.UsingNamespaces.Add("System.Text");
            sc.UsingNamespaces.Add("System.Collections.Generic");
            sc.UsingNamespaces.Add("HigLabo.DbSharp");

            var ns = new Namespace(namespaceName);
            ns.Classes.Add(CreateClass());
            sc.Namespaces.Add(ns);

            return new SourceCodeFile(Path.Combine(directoryPath, t.Name + ".cs"), sc);
        }
   
        public Class CreateClass()
        {
            var t = this.Table;
            var c = new Class(AccessModifier.Public, t.Name);

            c.BaseClass = new TypeName(String.Format("Table<{0}.Record>", t.Name));
            c.Modifier.Partial = true;

            c.Fields.Add(CreateNameConstField());
            c.Properties.Add(CreateStoredProcedureNameProperty());

            c.Methods.Add(CreateSelectAllMethod());
            c.Methods.Add(CreateSelectByPrimaryKeyMethod());
            c.Methods.Add(CreateSelectByPrimaryKeyOrNullMethod());
            c.Methods.Add(CreateInsertMethod());
            c.Methods.Add(CreateUpdateMethod());
            c.Methods.Add(CreateDeleteMethod());
            c.Methods.Add(CreateDeleteByValueMethod());
            c.Methods.Add(CreateInsertStoredProcedureMethod());
            c.Methods.Add(CreateUpdateStoredProcedureMethod());
            c.Methods.Add(CreateDeleteStoredProcedureMethod());

            return c;
        }

        public Field CreateNameConstField()
        {
            var f = new Field("String", "Name");
            f.Modifier.AccessModifier = AccessModifier.Public;
            f.Modifier.Const = true;
            f.Initializer = String.Format("\"{0}\"", this.Table.Name);
            return f;
        }
        public Property CreateStoredProcedureNameProperty()
        {
            var t = this.Table;
            var p = new Property("String", "TableName");
            p.Modifier.Polymophism = MethodPolymophism.Override;
            p.Get.Body.Add(SourceCodeLanguage.CSharp, "return {0}.Name;", t.Name);
            p.Set = null;
            return p;
        }

        public Method CreateSelectAllMethod()
        {
            var t = this.Table;
            Method md = new Method(MethodAccessModifier.Public, "SelectAll");
            md.Modifier.Polymophism = MethodPolymophism.Override;
            md.ReturnTypeName = new TypeName("List<Record>");

            md.Body.Add(SourceCodeLanguage.CSharp, "var sp = new {0}SelectAll();", t.Name);
            md.Body.Add(SourceCodeLanguage.CSharp, "var l = new List<Record>();");
            var cb = md.Body.Add(SourceCodeLanguage.CSharp, "foreach (var rs in sp.GetResultSets())");
            {
                cb.CurlyBracket = true;
                cb.CodeBlocks.Add(new CodeBlock(SourceCodeLanguage.CSharp, "var r = new Record(rs);"));
                cb.CodeBlocks.Add(new CodeBlock(SourceCodeLanguage.CSharp, "r.SetOldRecordProperty();"));
                cb.CodeBlocks.Add(new CodeBlock(SourceCodeLanguage.CSharp, "l.Add(r);"));
            }
            md.Body.Add(SourceCodeLanguage.CSharp, "return l;");

            return md;
        }
        public Method CreateSelectByPrimaryKeyMethod()
        {
            var t = this.Table;
            Method md = new Method(MethodAccessModifier.Public, "SelectByPrimaryKey");
            StringBuilder sb = new StringBuilder();
            Boolean isFirst = true;

            md.ReturnTypeName = new TypeName("Record");
            foreach (var column in t.GetPrimaryKeyColumns())
            {
                var pName = this.ToCamelCase(column.Name);

                md.Parameters.Add(new MethodParameter(column.GetClassName(), pName));

                if (isFirst == false)
                {
                    sb.Append(", ");
                }
                sb.Append(pName);
                isFirst = false;
            }
            md.Body.Add(SourceCodeLanguage.CSharp, "var r = this.SelectByPrimaryKeyOrNull({0});", sb.ToString());
            md.Body.Add(SourceCodeLanguage.CSharp, "if (r == null) throw new TableRecordNotFoundException();");
            md.Body.Add(SourceCodeLanguage.CSharp, "return r;");

            return md;
        }
        public Method CreateSelectByPrimaryKeyOrNullMethod()
        {
            var t = this.Table;
            Method md = new Method(MethodAccessModifier.Public, "SelectByPrimaryKeyOrNull");
            md.ReturnTypeName = new TypeName("Record");

            md.Body.Add(SourceCodeLanguage.CSharp, "var sp = new {0}SelectBy_PrimaryKey();", t.Name);
            foreach (var column in t.GetPrimaryKeyColumns())
            {
                var pName = this.ToCamelCase(column.Name);
                md.Parameters.Add(new MethodParameter(column.GetClassName(), pName));
                md.Body.Add(SourceCodeLanguage.CSharp, "sp.PK_{0} = {1};", column.Name, pName);
            }
            md.Body.Add(SourceCodeLanguage.CSharp, "var rs = sp.GetResultSet();", t.Name);
            md.Body.Add(SourceCodeLanguage.CSharp, "if (rs == null) return null;");
            md.Body.Add(SourceCodeLanguage.CSharp, "var r = new Record(rs);");
            md.Body.Add(SourceCodeLanguage.CSharp, "r.SetOldRecordProperty();");
            md.Body.Add(SourceCodeLanguage.CSharp, "return r;");
            return md;
        }
        public Method CreateInsertMethod()
        {
            var t = this.Table;
            Method md = new Method(MethodAccessModifier.Public, "Insert");
            md.Modifier.Polymophism = MethodPolymophism.Override;
            md.ReturnTypeName = new TypeName("Int32");
            md.Parameters.Add(new MethodParameter("Record", "record"));

            md.Body.Add(SourceCodeLanguage.CSharp, "if (record == null) throw new ArgumentNullException(\"record\");");
            md.Body.Add(SourceCodeLanguage.CSharp, "var sp = this.CreateInsertStoredProcedure(record);");
            md.Body.Add(SourceCodeLanguage.CSharp, "var x = sp.ExecuteNonQuery();");
            md.Body.Add(SourceCodeLanguage.CSharp, "if (x == 1) record.SetOldRecordProperty();");
            md.Body.Add(SourceCodeLanguage.CSharp, "return x;");
            return md;
        }
        public Method CreateUpdateMethod()
        {
            var t = this.Table;
            Method md = new Method(MethodAccessModifier.Public, "Update");
            md.Modifier.Polymophism = MethodPolymophism.Override;
            md.ReturnTypeName = new TypeName("Int32");
            md.Parameters.Add(new MethodParameter("Record", "record"));

            md.Body.Add(SourceCodeLanguage.CSharp, "if (record == null) throw new ArgumentNullException(\"record\");");
            md.Body.Add(SourceCodeLanguage.CSharp, "var sp = this.CreateUpdateStoredProcedure(record);");
            md.Body.Add(SourceCodeLanguage.CSharp, "var x = sp.ExecuteNonQuery();");
            md.Body.Add(SourceCodeLanguage.CSharp, "if (x == 1) record.SetOldRecordProperty();");
            md.Body.Add(SourceCodeLanguage.CSharp, "return x;");

            return md;
        }
        public Method CreateDeleteMethod()
        {
            var t = this.Table;
            Method md = new Method(MethodAccessModifier.Public, "Delete");
            md.Modifier.Polymophism = MethodPolymophism.Override;
            md.ReturnTypeName = new TypeName("Int32");
            md.Parameters.Add(new MethodParameter("Record", "record"));

            md.Body.Add(SourceCodeLanguage.CSharp, "if (record == null) throw new ArgumentNullException(\"record\");");
            md.Body.Add(SourceCodeLanguage.CSharp, "var sp = this.CreateDeleteStoredProcedure(record);");
            md.Body.Add(SourceCodeLanguage.CSharp, "var x = sp.ExecuteNonQuery();");
            md.Body.Add(SourceCodeLanguage.CSharp, "if (x == 1) record.SetOldRecordProperty();");
            md.Body.Add(SourceCodeLanguage.CSharp, "return x;");

            return md;
        }
        public Method CreateDeleteByValueMethod()
        {
            var t = this.Table;
            Method md = new Method(MethodAccessModifier.Public, "Delete");
            md.ReturnTypeName = new TypeName("Int32");

            md.Body.Add(SourceCodeLanguage.CSharp, "var sp = new {0}Delete();", t.Name);
            md.Body.Add(SourceCodeLanguage.CSharp, "((ITransaction)sp).TransactionKey = this.TransactionKey;");
            foreach (var column in t.GetPrimaryKeyColumns())
            {
                var pName = this.ToCamelCase(column.Name);

                md.Parameters.Add(new MethodParameter(column.GetClassName(), pName));
                md.Body.Add(SourceCodeLanguage.CSharp, "sp.PK_{0} = {1};", column.Name, pName);
            }
            md.Body.Add(SourceCodeLanguage.CSharp, "return sp.ExecuteNonQuery();");

            return md;
        }
        public Method CreateInsertStoredProcedureMethod()
        {
            var t = this.Table;
            Method md = new Method(MethodAccessModifier.Public, "CreateInsertStoredProcedure");
            md.ReturnTypeName = new TypeName(String.Format("{0}Insert", t.Name));
            md.Parameters.Add(new MethodParameter("Record", "record"));

            md.Body.AddRange(CreateInsertStoredProcedureMethodBody());

            return md;
        }
        public IEnumerable<CodeBlock> CreateInsertStoredProcedureMethodBody()
        {
            var t = this.Table;
            yield return new CodeBlock(SourceCodeLanguage.CSharp, "var sp = new {0}Insert();", t.Name);
            yield return new CodeBlock(SourceCodeLanguage.CSharp, "((ITransaction)sp).TransactionKey = this.TransactionKey;");
            foreach (var column in t.Columns)
            {
                yield return new CodeBlock(SourceCodeLanguage.CSharp, "sp.{0} = record.{0};", column.Name);
            }
            yield return new CodeBlock(SourceCodeLanguage.CSharp, "return sp;");
        }
        public Method CreateUpdateStoredProcedureMethod()
        {
            var t = this.Table;
            Method md = new Method(MethodAccessModifier.Public, "CreateUpdateStoredProcedure");
            md.ReturnTypeName = new TypeName(String.Format("{0}Update", t.Name));
            md.Parameters.Add(new MethodParameter("Record", "record"));

            md.Body.AddRange(CreateUpdateStoredProcedureMethodBody());

            return md;
        }
        public IEnumerable<CodeBlock> CreateUpdateStoredProcedureMethodBody()
        {
            var t = this.Table;

            yield return new CodeBlock(SourceCodeLanguage.CSharp, "if (record.OldRecord == null) throw new OldRecordIsNullException();");
            
            yield return new CodeBlock(SourceCodeLanguage.CSharp, "var sp = new {0}Update();", t.Name);
            yield return new CodeBlock(SourceCodeLanguage.CSharp, "((ITransaction)sp).TransactionKey = this.TransactionKey;");
            foreach (var column in t.Columns)
            {
                yield return new CodeBlock(SourceCodeLanguage.CSharp, "sp.{0} = record.{0};", column.Name);
            }
            foreach (var column in t.GetPrimaryKeyColumns())
            {
                yield return new CodeBlock(SourceCodeLanguage.CSharp, "sp.PK_{0} = record.OldRecord.{0};", column.Name);
            }
            yield return new CodeBlock(SourceCodeLanguage.CSharp, "return sp;");
        }
        public Method CreateDeleteStoredProcedureMethod()
        {
            var t = this.Table;
            Method md = new Method(MethodAccessModifier.Public, "CreateDeleteStoredProcedure");
            md.ReturnTypeName = new TypeName(String.Format("{0}Delete", t.Name));
            md.Parameters.Add(new MethodParameter("Record", "record"));

            md.Body.AddRange(CreateDeleteStoredProcedureMethodBody());

            return md;
        }
        public IEnumerable<CodeBlock> CreateDeleteStoredProcedureMethodBody()
        {
            var t = this.Table;

            yield return new CodeBlock(SourceCodeLanguage.CSharp, "if (record.OldRecord == null) throw new OldRecordIsNullException();");

            yield return new CodeBlock(SourceCodeLanguage.CSharp, "var sp = new {0}Delete();", t.Name);
            yield return new CodeBlock(SourceCodeLanguage.CSharp, "((ITransaction)sp).TransactionKey = this.TransactionKey;");
            foreach (var column in t.GetPrimaryKeyColumns())
            {
                yield return new CodeBlock(SourceCodeLanguage.CSharp, "sp.PK_{0} = record.OldRecord.{0};", column.Name);
            }
            yield return new CodeBlock(SourceCodeLanguage.CSharp, "return sp;");
        }

        private String ToCamelCase(String value)
        {
            if (String.IsNullOrEmpty(value) == true) return value;
            if (value.Length == 2) return value.ToLower();
            return value.Substring(0, 1).ToLower() + value.Substring(1, value.Length - 1);
        }
    }
}
