﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using HigLabo.CodeGenerator;
using System.IO;

namespace HigLabo.DbSharp.CodeGenerator
{
    public class TableRecordClassSourceCodeFileFactory : ClassSourceCodeFileFactory
    {
        public Table Table { get; private set; }

        public TableRecordClassSourceCodeFileFactory(Table table)
        {
            this.Table = table;
        }

        public override SourceCodeFile Create(String directoryPath, String namespaceName)
        {
            var t = this.Table;
            var sc = new SourceCode();

            sc.UsingNamespaces.Add("System");
            sc.UsingNamespaces.Add("System.Data");
            sc.UsingNamespaces.Add("System.Text");
            sc.UsingNamespaces.Add("HigLabo.DbSharp");

            var ns = new Namespace(namespaceName);
            ns.Classes.Add(CreateClass());
            sc.Namespaces.Add(ns);

            return new SourceCodeFile(Path.Combine(directoryPath, t.Name + ".Record.cs"), sc);
        }
        public Class CreateClass()
        {
            var t = this.Table;
            var c = new Class(AccessModifier.Public, t.Name);

            c.Modifier.Partial = true;

            c.Classes.Add(this.CreateRecordClass());

            return c;
        }
        
        public Class CreateRecordClass()
        {
            var t = this.Table;
            var c = new Class(AccessModifier.Public, "Record");

            c.Modifier.Partial = true;
            c.BaseClass = new TypeName("TableRecord");
            c.ImplementInterfaces.Add(new TypeName("IRecord"));

            c.Properties.Add(CreateOldRecordProperty());

            c.Constructors.Add(CreateDefaultConstructor());
            c.Constructors.Add(CreateConstructor());

            c.Methods.Add(CreateGetTableNameMethod());
            c.Methods.Add(new Method(MethodAccessModifier.Partial, "ConstructorExecuted"));
            c.Methods.Add(CreateSetOldRecordPropertyMethod());
            c.Methods.Add(CreateSetPropertyMethod());
            c.Methods.Add(CreateIsChangedMethod());
            c.Methods.Add(CreateCompareAllColumnMethod());

            ClassSourceCodeFileFactory.AddPropertyAndField(c, t.Columns);

            return c;
        }

        public Method CreateGetTableNameMethod()
        {
            var t = this.Table;
            var md = new Method(MethodAccessModifier.Public, "GetTableName");
            md.Modifier.Polymophism = MethodPolymophism.Override;
            md.ReturnTypeName = new TypeName("String");
            md.Body.Add(SourceCodeLanguage.CSharp, String.Format("return \"{0}\";", t.Name)); 

            return md;
        }
        public Property CreateOldRecordProperty()
        {
            var t = this.Table;
            var p = new Property("Record", "OldRecord");
            p.Get.IsAutomaticProperty = true;
            p.Set.IsAutomaticProperty = true;
            p.Set.Modifier = AccessModifier.Private;
            return p;
        }

        public Constructor CreateDefaultConstructor()
        {
            var t = this.Table;
            Constructor ct = new Constructor(AccessModifier.Public, "Record");

            ct.Body.Add(SourceCodeLanguage.CSharp, "ConstructorExecuted();");
            return ct;
        }
        public Constructor CreateConstructor()
        {
            var t = this.Table;
            Constructor ct = new Constructor(AccessModifier.Public, "Record");

            ct.Parameters.Add(new MethodParameter("IRecord", "record"));
            ct.Body.Add(SourceCodeLanguage.CSharp, "this.SetProperty(record);");
            ct.Body.Add(SourceCodeLanguage.CSharp, "ConstructorExecuted();");
            return ct;
        }

        public Method CreateSetPropertyMethod()
        {
            var t = this.Table;
            Method md = new Method(MethodAccessModifier.Public, "SetProperty");
            md.Parameters.Add(new MethodParameter("IRecord", "record"));
            md.ReturnTypeName = new TypeName("void");
            md.Body.Add(SourceCodeLanguage.CSharp, "var r = record;");
            foreach (var column in t.Columns)
            {
                md.Body.Add(SourceCodeLanguage.CSharp, "this.{0} = r.{0};", column.Name);
            }
            return md;
        }
        public Method CreateSetOldRecordPropertyMethod()
        {
            var t = this.Table;
            Method md = new Method(MethodAccessModifier.Public, "SetOldRecordProperty");
            md.ReturnTypeName = new TypeName("void");
            md.Body.Add(SourceCodeLanguage.CSharp, "this.OldRecord = new Record(this);");
            return md;
        }
        public Method CreateIsChangedMethod()
        {
            var t = this.Table;
            Method md = new Method(MethodAccessModifier.Public, "IsChanged");
            md.ReturnTypeName = new TypeName("Boolean");
            md.Body.Add(SourceCodeLanguage.CSharp, "if (this.OldRecord == null) throw new InvalidOperationException(\"You must call SetOldRecordProperty method before call IsChanged method.\");");
            md.Body.Add(SourceCodeLanguage.CSharp, "return this.CompareAllColumn(this.OldRecord);");
            return md;
        }
        public Method CreateCompareAllColumnMethod()
        {
            var t = this.Table;
            Method md = new Method(MethodAccessModifier.Public, "CompareAllColumn");
            md.Parameters.Add(new MethodParameter("IRecord", "record"));
            md.ReturnTypeName = new TypeName("Boolean");
            md.Body.Add(SourceCodeLanguage.CSharp, "if (record == null) throw new ArgumentNullException(\"record\");");
            md.Body.Add(SourceCodeLanguage.CSharp, "var r = record;");

            StringBuilder sb = new StringBuilder();
            for (int i = 0; i < t.Columns.Count; i++)
            {
                var column = t.Columns[i];
                sb.Clear();
                if (i == 0)
                {
                    sb.Append("return ");
                }
                if (i < t.Columns.Count - 1)
                {
                    sb.AppendFormat("this.{0} == r.{0} && ", column.Name);
                }
                else
                {
                    sb.AppendFormat("this.{0} == r.{0};", column.Name);
                }
                md.Body.Add(SourceCodeLanguage.CSharp, sb.ToString());
            }
            return md;
        }
    }
}
