﻿using HigLabo.Data;
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

namespace HigLabo.DbSharp.CodeGenerator
{
    public abstract class DatabaseSchemaReader
    {
        public String ConnectionString { get; protected set; }
        public static DatabaseSchemaReader Create(DatabaseServer databaseServer, String connectionString)
        {
            switch (databaseServer)
            {
                case DatabaseServer.SqlServer: return new SqlServerDatabaseSchemaReader(connectionString);
                case DatabaseServer.Oracle: throw new NotImplementedException();
                case DatabaseServer.MySql: return new MySqlDatabaseSchemaReader(connectionString);
                case DatabaseServer.PostgreSql: throw new NotImplementedException();
                default: throw new InvalidOperationException();
            }
        }
        public abstract IDatabase CreateDatabase();
        public abstract List<DatabaseObject> GetUserTableList();
        public abstract Table GetTable(String name);
        public abstract List<Column> GetColumns(String tableName);
        public abstract List<DatabaseObject> GetUserStoredProcedureList();
        public abstract StoredProcedure GetStoredProcedure(String name);
        public abstract List<SqlParameter> GetParameters(String storedProcedureName);
        public abstract List<StoredProcedureParameter> GetResultSets(StoredProcedure sp, IEnumerable<SqlParameter> parameters);
        public abstract List<StructuredDataTableColumn> GetStructuredDataTableColumnList(String tableName);
        public abstract Boolean ExistStoredProcedure(String name);

        public String CreateQueryOfTableNameSelectAll(Table table)
        {
            return String.Format("create procedure {0}SelectAll as " + Environment.NewLine
                + "select * from [{0}] with(nolock)", table.Name);
        }
        public String CreateQueryOfTableNameSelectBy_PrimaryKey(Table table)
        {
            StringBuilder sb = new StringBuilder(32);

            sb.AppendFormat("create procedure {0}SelectBy_PrimaryKey ", table.Name);
            sb.Append("(");
            sb.Append(CreateText(table.GetPrimaryKeyColumns(), column => String.Format("@PK_{0}", column.GetDeclareParameterText()), ",", true));
            sb.AppendLine(") as");
            sb.AppendLine();

            sb.AppendFormat("select * from [{0}] with(nolock)", table.Name);
            sb.AppendLine();
            sb.Append("where ");
            sb.Append(CreateText(table.GetPrimaryKeyColumns(), column => String.Format("[{0}] = @PK_{0}", column.Name), "and ", true));

            return sb.ToString();
        }
        public String CreateQueryOfTableNameInsert(Table table)
        {
            StringBuilder sb = new StringBuilder(32);
            List<Column> outputColumns = new List<Column>();
            List<Column> insertedColumns = new List<Column>();

            foreach (var column in table.Columns)
            {
                if (column.IsServerAutomaticallyInsertValueColumn() == true)
                {
                    if (column.IsRowGuidCol == true ||
                        column.IsIdentity == true)
                    {
                        insertedColumns.Add(column);
                    }
                    else
                    {
                        outputColumns.Add(column);
                    }
                }
            }
            
            sb.AppendFormat("CREATE PROCEDURE {0}Insert", table.Name);
            sb.AppendLine();
            sb.Append("(");
            sb.Append(CreateText(table.Columns, column =>
            {
                if (column.IsServerAutomaticallyInsertValueColumn() == true)
                {
                    return "@" + column.GetDeclareParameterText() + " output";
                }
                else
                {
                    return "@" + column.GetDeclareParameterText();
                }
            }, ",", true));
            sb.AppendLine(") as");
            sb.AppendLine();

            if (insertedColumns.Count > 0)
            {
                sb.Append("declare @InsertedTable table(");
                sb.Append(this.CreateText(insertedColumns, column => column.GetDeclareParameterText(), ", ", false));
                sb.AppendLine(");");
                sb.AppendLine();
            }
            sb.AppendFormat("insert into [{0}]", table.Name);
            sb.AppendLine();

            sb.Append("(");
            sb.Append(CreateText(table.GetColumns(null, false), column => String.Format("[{0}]", column.Name), ",", true));
            sb.AppendLine(")");
            if (insertedColumns.Count > 0)
            {
                sb.Append("output ");
                sb.Append(CreateText(insertedColumns, column => String.Format("inserted.{0}", column.Name), ", ", false));
                sb.AppendFormat(" into @InsertedTable");
                sb.AppendLine();
            }
            sb.AppendLine("values");

            sb.Append("(");
            sb.Append(CreateText(table.GetColumns(null, false), column => "@" + column.Name, ",", true));
            sb.AppendLine(")");

            if (outputColumns.Count > 0)
            {
                sb.AppendLine();
                sb.Append("select ");
                sb.Append(CreateText(outputColumns, column => String.Format("@{0} = {0}", column.Name), ", ", false));
                sb.AppendFormat(" from {0} with(nolock)", table.Name);
                sb.AppendLine();
                sb.AppendLine(this.CreateWhereQuery(table.GetPrimaryKeyColumns(), false));
            }
            if (insertedColumns.Count > 0)
            {
                sb.Append("select ");
                sb.Append(CreateText(insertedColumns, column => String.Format("@{0} = {0}", column.Name), ", ", false));
                sb.Append(" from @InsertedTable");
                sb.AppendLine();
            }

            return sb.ToString();
        }
        public String CreateQueryOfTableNameUpdate(Table table)
        {
            StringBuilder sb = new StringBuilder(32);
            List<Column> columns = null;
            List<Column> outputColumns = new List<Column>();
            String whereQuery = this.CreateWhereQuery(table.GetPrimaryKeyColumns(), true);

            foreach (var column in table.Columns)
            {
                if (column.CanUpdateValueColumn() == false)
                {
                    outputColumns.Add(column);
                }
            }

            sb.AppendFormat("create procedure {0}Update", table.Name);
            sb.AppendLine();
            sb.Append("(");
            sb.Append(CreateText(table.Columns, column => {
                if (column.CanUpdateValueColumn() == true)
                {
                    return "@" + column.GetDeclareParameterText();
                }
                else
                {
                    return "@" + column.GetDeclareParameterText() + " output";
                }
            }, ",", true));
            sb.Append(",");
            sb.Append(CreateText(table.GetPrimaryKeyColumns(), column => String.Format("@PK_{0}", column.GetDeclareParameterText()), ",", true));
            sb.AppendLine(") as");
            sb.AppendLine();

            sb.AppendLine("if(");
            sb.Append(CreateText(table.GetPrimaryKeyColumns(), column => String.Format("@{0} = @PK_{0}", column.Name), "and ", true));
            sb.AppendLine(")");

            columns = table.GetColumns(false, false).ToList();
            sb.AppendLine("begin");
            if (columns.Count == 0)
            {
                sb.AppendFormat("Set @{0} = @{0}", table.Columns[0].Name);
                sb.AppendLine();
            }
            else if (columns.Count > 0)
            {
                sb.AppendFormat("update [{0}] set ", table.Name);
                sb.AppendLine();

                sb.Append(CreateText(columns, column => String.Format("[{0}] = @{0}", column.Name), ",", true));
                sb.AppendLine();
                sb.Append(whereQuery);
            }
            sb.AppendLine("end");

            sb.AppendLine("else");

            sb.AppendLine("begin");
            sb.AppendFormat("update [{0}] set ", table.Name);
            sb.AppendLine();

            sb.Append(CreateText(table.Columns.Where(column => column.CanUpdateValueColumn()), column => String.Format("[{0}] = @{0}", column.Name), ",", true));
            sb.AppendLine();
            sb.Append(whereQuery);
            sb.AppendLine("end");

            if (outputColumns.Count > 0)
            {
                sb.AppendLine();
                sb.Append("select ");
                sb.Append(CreateText(outputColumns, column => String.Format("@{0} = {0}", column.Name), ", ", false));
                sb.AppendFormat(" from {0} with(nolock)", table.Name);
                sb.AppendLine();
                sb.AppendLine(this.CreateWhereQuery(table.GetPrimaryKeyColumns(), false));
            }

            return sb.ToString();
        }
        public String CreateQueryOfTableNameDelete(Table table)
        {
            StringBuilder sb = new StringBuilder(32);
            String whereQuery = this.CreateWhereQuery(table.GetPrimaryKeyColumns(), true);

            sb.AppendFormat("create procedure {0}Delete", table.Name);
            sb.Append("(");
            sb.Append(CreateText(table.GetPrimaryKeyColumns(), column => String.Format("@PK_{0}", column.GetDeclareParameterText()), ",", true));
            sb.AppendLine(") as");
            sb.AppendLine();

            sb.AppendFormat("delete [{0}]", table.Name);
            sb.AppendLine();
            sb.Append(whereQuery);

            return sb.ToString();
        }
        private String CreateWhereQuery(IEnumerable<Column> columns, Boolean addPK_Prefix)
        {
            StringBuilder sb = new StringBuilder(32);

            sb.Append("where ");
            sb.Append(CreateText(columns, column =>
            {
                if (addPK_Prefix == true)
                {
                    return String.Format("[{0}] = @PK_{0}", column.Name);
                }
                return String.Format("[{0}] = @{0}", column.Name);
            }, "and ", true));
            return sb.ToString();
        }
        private String CreateText<T>(IEnumerable<T> values, Func<T, String> selector, String separator, Boolean newLine)
        {
            StringBuilder sb = new StringBuilder();
            Boolean isFirst = true;
            foreach (var item in values)
            {
                if (isFirst == false)
                {
                    sb.Append(separator);
                }
                sb.Append(selector(item));
                if (newLine == true)
                {
                    sb.AppendLine();
                }
                isFirst = false;
            }
            return sb.ToString();
        }
    }
}
