﻿using HigLabo.Data;
using System;
using System.Collections.Generic;
using System.Data;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

namespace HigLabo.DbSharp
{
	/// <summary>
	/// 
	/// </summary>
    public class DatabaseContext : IDisposable
    {
        /// <summary>
        /// 
        /// </summary>
        public static event EventHandler<EventArgs> TransactionStarted;
     
        private static Func<IDatabase> _DefaultDatabaseFactory = null;
        /// <summary>
        /// 
        /// </summary>
        [ThreadStatic()]
        // Thread固有なのでロックは不要。
        // ここでnewするとこのスレッドだけしか値がセットされないのでGetDatabasesメソッドでインスタンスをセットする。
        private static Dictionary<String, DatabaseContext> _Contexts = null;
        /// <summary>
        /// 
        /// </summary>
        public IDatabase Database { get; private set; }
        /// <summary>
        /// 
        /// </summary>
        public String Key { get; private set; }
        /// <summary>
        /// 
        /// </summary>
        public DatabaseContext()
        {
            this.Initialize("", CreateDatabase(), null);
        }
        /// <summary>
        /// 
        /// </summary>
        /// <param name="key"></param>
        public DatabaseContext(String key)
        {
            this.Initialize(key, CreateDatabase(), null);
        }
        /// <summary>
        /// 
        /// </summary>
        /// <param name="key"></param>
        public DatabaseContext(String key, IDatabase database)
        {
            this.Initialize(key, database, null);
        }
        /// <summary>
		/// 
		/// </summary>
		/// <param name="database"></param>
		/// <param name="isolationLevel"></param>
        public DatabaseContext(String key, IsolationLevel isolationLevel)
        {
            var db = CreateDatabase();
            this.Initialize(key, db, isolationLevel);
        }
        /// <summary>
        /// 
        /// </summary>
        /// <param name="key"></param>
        /// <param name="isolationLevel"></param>
        public DatabaseContext(String key, IDatabase database, IsolationLevel isolationLevel)
        {
            this.Initialize(key, database, isolationLevel);
        }
        private void Initialize(String key, IDatabase database, IsolationLevel? isolationLevel)
        {
            this.Key = key;
            this.Database = database;
            DatabaseContext.SetDatabaseContext(this.Key, this);
            if (isolationLevel.HasValue == true)
            {
                this.BeginTransaction(isolationLevel.Value);
            }
        }
        /// <summary>
        /// 
        /// </summary>
        public void Open()
        {
            IDatabase db = DatabaseContext.GetDatabase(this.Key);
            if (db == null)
            { throw new ObjectDisposedException("DatabaseContext.Database"); }
            db.Open();
        }
        /// <summary>
        /// 
        /// </summary>
        public void Close()
        {
            IDatabase db = DatabaseContext.GetDatabase(this.Key);
            if (db == null)
            { throw new ObjectDisposedException("DatabaseContext.Database"); }
            db.Close();
        }
        /// <summary>
        /// 
        /// </summary>
        /// <param name="isolationLevel"></param>
        public void BeginTransaction(IsolationLevel isolationLevel)
        {
            IDatabase db = DatabaseContext.GetDatabase(this.Key);
            if (db == null) { throw new InvalidOperationException(); }
            db.BeginTransaction(isolationLevel);
            DatabaseContext.OnTransactionStarted(new EventArgs());
        }
        /// <summary>
        /// 
        /// </summary>
        public void CommitTransaction()
        {
            IDatabase db = DatabaseContext.GetDatabase(this.Key);
            if (db.OnTransaction == false)
            {
                throw new InvalidOperationException("Transaction does not start or finished."
                  + "Please call BeginTransaction method of DatabaseContext object before call this method."
                  + "Or ensure call once CommitTransaction or RollbackTransaction.");
            }
            db.CommitTransaction();
        }
        /// <summary>
        /// 
        /// </summary>
        public void RollbackTransaction()
        {
            IDatabase db = DatabaseContext.GetDatabase(this.Key);
            if (db.OnTransaction == false)
            {
                throw new InvalidOperationException("Transaction does not start or finished."
                  + "Please call BeginTransaction method of DatabaseContext object before call this method."
                  + "Or ensure call once CommitTransaction or RollbackTransaction.");
            }
            db.RollBackTransaction();
        }
        /// <summary>
        /// 
        /// </summary>
        /// <param name="func"></param>
        public static void SetCreateDatabaseFunc(Func<IDatabase> func)
        {
            _DefaultDatabaseFactory = func;
        }
        private static Dictionary<String, DatabaseContext> GetDatabaseContextList()
        {
            if (DatabaseContext._Contexts == null)
            {
                DatabaseContext._Contexts = new Dictionary<String, DatabaseContext>();
            }
            return DatabaseContext._Contexts;
        }
        /// <summary>
        /// 
        /// </summary>
        /// <returns></returns>
        public static IDatabase CreateDatabase()
        {
            if (_DefaultDatabaseFactory == null)
            {
                throw new InvalidOperationException("You must set DefaultDatabaseFactory property of DatabaseContext class."
                    + "Please call SetDefaultDatabaseFactory method.");
            }
            return _DefaultDatabaseFactory();
        }
        /// <summary>
        /// 
        /// </summary>
        /// <param name="key"></param>
        /// <returns></returns>
        internal static IDatabase GetDatabase(String key)
        {
            if (key == null)
            {
                return DatabaseContext.CreateDatabase();
            }
            var dcs = DatabaseContext.GetDatabaseContextList();
            DatabaseContext dc = null;
            dcs.TryGetValue(key, out dc);
            if (dc == null)
            {
                return DatabaseContext.CreateDatabase();
            }
            return dc.Database;
        }
        /// <summary>
        /// 
        /// </summary>
        /// <param name="key"></param>
        /// <param name="database"></param>
        private static void SetDatabaseContext(String key, DatabaseContext database)
        {
            var db = DatabaseContext.GetDatabaseContextList();
            if (db.ContainsKey(key) == true) throw new TransactionKeyAlreadyUsedException();
            db[key] = database;
        }
        /// <summary>
        /// 
        /// </summary>
        /// <param name="e"></param>
        protected static void OnTransactionStarted(EventArgs e)
        {
            var eh = DatabaseContext.TransactionStarted;
            if (eh != null)
            {
                eh(null, e);
            }
        }
        /// <summary>
        /// 
        /// </summary>
        public void Dispose()
        {
            IDatabase db = DatabaseContext.GetDatabase(this.Key);

            if (db != null)
            {
                if (DatabaseContext._Contexts.ContainsKey(this.Key) == true)
                {
                    DatabaseContext._Contexts[this.Key] = null;
                }
                db.Dispose();
            }
        }
    }
}
